from typing import Callable, List, Optional, Tuple

import numpy as np
from image_keras.supports.tuple_op import (
    tuple_add,
    tuple_divide_int,
    tuple_element_wise_add,
    tuple_element_wise_divide_int,
    tuple_element_wise_subtract,
    tuple_multiply,
)


def add_zero_padding(
    cv2_image: np.ndarray, padding_height_width: Tuple[int, int]
) -> Optional[np.ndarray]:
    """
    Add zero padding for open cv image. (grayscale image (w, h) or rgb image (w, h, c).)

    Parameters
    ----------
    cv2_image : np.ndarray
        Open cv image to add zero padding.
    padding_height_width : Tuple[int, int]
        Zero padding width and height.

    Returns
    -------
    Optional[np.ndarray]
        Zero padded image. If not image, it will return None.
    """
    if len(cv2_image.shape) == 3:
        return np.pad(
            cv2_image,
            (
                (padding_height_width[0], padding_height_width[0]),
                (padding_height_width[1], padding_height_width[1]),
                (0, 0),
            ),
            "constant",
            constant_values=0,
        )
    elif len(cv2_image.shape) == 2:
        return np.pad(
            cv2_image,
            (
                (padding_height_width[0], padding_height_width[0]),
                (padding_height_width[1], padding_height_width[1]),
            ),
            "constant",
            constant_values=0,
        )
    else:
        return None


def num_total_index_for_slice(
    img_height_width: Tuple[int, int],
    tile_height_width: Tuple[int, int],
    padding_height_width: Tuple[int, int],
    stride_height_width: Tuple[int, int],
    add_same_padding: bool = False,
    discard_horizontal_overflow: bool = True,
    discard_vertical_overflow: bool = True,
) -> Tuple[int, int]:
    base_wh = tuple_element_wise_subtract(img_height_width, tile_height_width)
    if add_same_padding:
        base_wh = tuple_element_wise_add(
            base_wh, tuple_multiply(padding_height_width, 2)
        )
    num_height_width_index = tuple_element_wise_divide_int(
        base_wh,
        stride_height_width,
    )
    num_height_width_index = tuple_add(num_height_width_index, 1)
    if not discard_vertical_overflow:
        num_height_width_index[0] = num_height_width_index[0] + 1
    if not discard_horizontal_overflow:
        num_height_width_index[1] = num_height_width_index[1] + 1
    return num_height_width_index


def slice_by_pixel_size(
    cv2_image: np.ndarray,
    tile_height_width: Tuple[int, int],
    inbox_height_width: Tuple[int, int],
    stride_height_width: Tuple[int, int],
    add_same_padding: bool = False,
    discard_horizontal_overflow: bool = True,
    discard_vertical_overflow: bool = True,
) -> List[List[np.ndarray]]:
    """
    Slice image.

    Parameters
    ----------
    cv2_image : np.ndarray
        [description]
    tile_height_width : Tuple[int, int]
        [description]
    inbox_height_width : Tuple[int, int]
        [description]
    stride_height_width : Tuple[int, int]
        [description]
    discard_horizontal_overflow : bool, optional
        [description], by default True
    discard_vertical_overflow : bool, optional
        [description], by default True

    Returns
    -------
    List[List[np.ndarray]]
        [description]

    Examples
    --------
    >>> image: np.ndarray = cv2.imread('/path/to/image_file.png')
    >>> slice_by_pixel_size(image, tile_height_width=(128, 128), inbox_height_width=(60, 60), stride_height_width=(60, 60))
    """
    padding_height_width = tuple_divide_int(
        tuple_element_wise_subtract(
            tile_height_width,
            inbox_height_width,
        ),
        2,
    )

    num_height_width_index = num_total_index_for_slice(
        cv2_image.shape[:2],
        tile_height_width,
        padding_height_width,
        stride_height_width,
        add_same_padding,
        discard_horizontal_overflow,
        discard_vertical_overflow,
    )

    if add_same_padding:
        padding_cv2_image = add_zero_padding(cv2_image, padding_height_width)
    else:
        padding_cv2_image = cv2_image

    result_block_box: List[List[np.ndarray]] = []
    for ri in range(num_height_width_index[0]):
        result_row: List[np.ndarray] = []
        from_height = stride_height_width[0] * ri
        to_height = stride_height_width[0] * ri + tile_height_width[0]

        for ci in range(num_height_width_index[1]):
            from_width = stride_height_width[1] * ci
            to_width = stride_height_width[1] * ci + tile_height_width[1]

            result_row.append(
                padding_cv2_image[from_height:to_height, from_width:to_width]
            )
        result_block_box.append(result_row)
    return result_block_box


def apply_each_slice(
    sliced_cv2_images: List[List[np.ndarray]],
    func: Callable[[int, int, np.ndarray], Optional[np.ndarray]],
) -> Tuple[int, int, List[List[np.ndarray]]]:
    i = 0
    j = 0
    result_block_box: List[List[np.ndarray]] = []
    for sliced_images_row in sliced_cv2_images:
        j = 0
        result_row: List[np.ndarray] = []
        for sliced_images_col in sliced_images_row:
            optional_r = func(i, j, sliced_images_col)
            if optional_r:
                result_row.append(optional_r)
            j = j + 1
        result_block_box.append(result_row)
        i = i + 1

    return i, j, result_block_box



import os
from typing import Tuple

import cv2
import numpy as np
from PIL import Image


def image_read(folder: str, name: str, verbose: int = 1) -> Tuple[str, np.ndarray]:
    print("Image read {}".format(name)) if verbose == 1 else None
    return name, cv2.imread(os.path.join(folder, name))


def image_read_16bit_grayscale(folder: str, name: str, verbose: int = 1):
    print("Image read {}".format(name)) if verbose == 1 else None
    return name, cv2.imread(os.path.join(folder, name), cv2.IMREAD_ANYDEPTH)


def image_write(folder: str, name: str, img: np.ndarray, verbose: int = 1) -> bool:
    print("Image write {}".format(name)) if verbose == 1 else None
    return cv2.imwrite(os.path.join(folder, name), img)


def to_pillow_image(cv2_image: np.ndarray, is_gray_image: bool = False) -> Image:
    cv2_image = (
        cv2.cvtColor(cv2_image, cv2.COLOR_BGR2RGB)
        if is_gray_image is not True
        else cv2_image
    )
    pil_image = Image.fromarray(cv2_image)
    return pil_image



import os
from typing import List, Tuple


def split_fullpath(full_path: str) -> Tuple[str, str, str]:
    """
    Split path to tuple.

    Parameters
    ----------
    full_path : str
        Full path to be decomposed.

    Returns
    -------
    Tuple[str, str, str]
        (full path, base filename, file extension) tuple.

    Examples
    --------
    >>> split_fullpath("/abc/test_resources/sample.png")
    ('/abc/test_resources', 'sample', '.png')

    >>> split_fullpath("/abc/test_resources")
    ('/abc/test_resources', None, None)
    """
    if os.path.isfile(full_path):
        _s = os.path.splitext(full_path)
        base_filename = os.path.basename(_s[0])
        base_ext = _s[1]
        pathname = os.path.dirname(_s[0])
        return pathname, base_filename, base_ext
    else:
        return full_path, None, None


def get_image_filenames(in_path: str) -> List[str]:
    """
    Get image filenames("jpg", "jpeg", "bmp", "png", "gif") in `in_path` folder.

    Parameters
    ----------
    in_path : str
        Folder name to get image filenames from.

    Returns
    -------
    List[str]
        Image filename list

    Examples
    --------
    >>> get_image_filenames("tests/test_resources)
    ['lenna.png', 'sample.png', 'bone.png', 'a.png']
    """
    included_extensions = ["jpg", "jpeg", "bmp", "png", "gif"]
    return [
        fn
        for fn in os.listdir(in_path)
        if any(fn.endswith(ext) for ext in included_extensions)
    ]



from typing import Tuple


def tuple_element_wise_add(a: Tuple, b: Tuple):
    return tuple(map(lambda el_1, el_2: el_1 + el_2, a, b))


def tuple_element_wise_subtract(a: Tuple, b: Tuple):
    return tuple(map(lambda el_1, el_2: el_1 - el_2, a, b))


def tuple_element_wise_multiply(a: Tuple, b: Tuple):
    return tuple(map(lambda el_1, el_2: el_1 * el_2, a, b))


def tuple_element_wise_divide(a: Tuple, b: Tuple):
    return tuple(map(lambda el_1, el_2: el_1 / el_2, a, b))


def tuple_element_wise_divide_int(a: Tuple, b: Tuple):
    return tuple(map(lambda el_1, el_2: el_1 // el_2, a, b))


def tuple_add(a: Tuple, b: int):
    return tuple(map(lambda el: el + b, a))


def tuple_subtract(a: Tuple, b: int):
    return tuple(map(lambda el: el - b, a))


def tuple_multiply(a: Tuple, b: int):
    return tuple(map(lambda el: el * b, a))


def tuple_divide(a: Tuple, b: int):
    return tuple(map(lambda el: el / b, a))


def tuple_divide_int(a: Tuple, b: int):
    return tuple(map(lambda el: el // b, a))



from typing import List, TypeVar

T = TypeVar("T")


def list_diff(f: List[T], t: List[T]) -> List[T]:
    return list(set(f) - set(t))


def list_intersection(f: List[T], t: List[T]) -> List[T]:
    return list(set(f).intersection(t))



from typing import Mapping, Optional, TypeVar

import toolz

K = TypeVar("K")
V = TypeVar("V")


def get_from_map(
    m: Mapping[K, V], key: K, default_value_optional: Optional[V] = None
) -> Optional[V]:
    return toolz.get_in([key], m, default_value_optional)



import glob
import os
import re
import shutil
from typing import Callable, List, Optional, Tuple

import toolz


def files_in_folder(
    folder_name: str, file_name_filter_optional: Optional[Callable[[str], bool]] = None
) -> List[str]:
    """
    Get files in folder without hidden file (starts with '.').

    Filtering using `file_name_filter_optional`.

    Args:
        folder_name (str): folder name
        file_name_filter_optional (Optional[Callable[[str], bool]]): optional filter

    Returns:
        File name lists
    """
    ff = (
        (lambda file_name: file_name == file_name)
        if file_name_filter_optional is None
        else (file_name_filter_optional)
    )
    return toolz.pipe(
        folder_name,
        os.listdir,
        toolz.curried.filter(lambda f: os.path.isfile(folder_name + "/" + f)),
        toolz.curried.filter(lambda file: not file.startswith(".")),
        toolz.curried.filter(ff),
        list,
        sorted,
    )


def create_folder_if_not_exist(folder_path: str) -> None:
    os.makedirs(folder_path) if not os.path.exists(folder_path) else None


def get_files_only_in_folder(folder_path: str) -> List[str]:
    files = files_in_folder(folder_path)
    files = list(
        filter(lambda el: os.path.isfile(os.path.join(folder_path, el)), files)
    )
    return files


def move_all_file_to_folder(from_folder: str, _target_folder: str) -> None:
    for file in get_files_only_in_folder(from_folder):
        shutil.move(os.path.join(from_folder, file), _target_folder)


def move_overwrite_all_file_to_folder(from_folder: str, _target_folder: str) -> None:
    cwd = os.getcwd()
    for file in get_files_only_in_folder(from_folder):
        shutil.move(
            os.path.join(cwd, from_folder, file),
            os.path.join(cwd, _target_folder, file),
        )


def copy_all_file_to_folder(from_folder: str, _target_folder: str) -> None:
    for file in get_files_only_in_folder(from_folder):
        shutil.copy2(os.path.join(from_folder, file), _target_folder)


def remove_files(starts_with_list: List[str], target_path: str) -> None:
    for _file_starts_with in starts_with_list:
        for _file in glob.glob("{}/{}*".format(target_path, _file_starts_with)):
            os.remove(_file)


def rename_files(
    sorted_list1: List[str], sorted_list2: List[str], target_path: str
) -> None:
    rename_files_from_list: List[Tuple[int, Tuple[str, str]]] = []
    _rev_current_file_names = sorted_list1
    for index, current_file_name in enumerate(sorted_list2):
        rename_files_from_list.append(
            (index, (current_file_name, _rev_current_file_names[index]))
        )
    for rename_files_for_prev_from_to in sorted(rename_files_from_list):
        for old_name in glob.glob(
            "{}/{}*".format(target_path, rename_files_for_prev_from_to[1][0])
        ):
            new_name = re.sub(
                r"(.*)(.{6}).*\.png",
                rename_files_for_prev_from_to[1][1] + r"\2.png",
                os.path.basename(old_name),
            )
            os.rename(old_name, os.path.join(target_path, new_name))
